from . import (
    circuit,
    noise,
    converter
)

from .circuit import (
    Operation,
    ImmutableOperation,
    State,
    Measurement,
    Channel,
    PureOperation,
    PureState,
    Unitary,
    PureMeas,
    ControlledOperation,
    Controlled,
    Diagonal,
    CompState,
    CompMeas,
    XGate,
    YGate,
    ZGate,
    TGate,
    HGate,
    SGate,
    IGate,
    SWAPGate,
    XHalfGate,
    YHalfGate,
    CNOTGate,
    Trace,
    XRotation,
    ZRotation,
    Circuit,
    ControlledCircuit,
    XXRotation,
    ZZRotation,
    SuperPosition,
    FourierMeas,
    FourierState,
    ZeroState,
    OneState,
    PlusState,
    MinusState,
    ZeroMeas,
    OneMeas,
    PlusMeas,
    MinusMeas,
    CZGate
)

from .noise import (
    Depolarization,
    Dephasing,
    AmplitudeDampling,
    add_noise
)

from .converter import (
    Converter
)



from acqdp.tensor_network.contraction import ContractionCost
import copy
import opt_einsum
import sys
import networkx
from networkx.algorithms import max_weight_matching, ancestors, descendants

sys.setrecursionlimit(10**4)

CONTRACTION_TREE_PARAMS = {
    'stem_size': 23,
    'connection': 4,
    'k': 14
}


class ContractionTree:
    """
    Contraction tree corresponding to a pairwise sequential contraction, with interfaces for branch flipping and merging.
    TODO: Merge with :class:`UndirectedContractionTree` for a unified interface for contraction trees.
    """
    @classmethod
    def from_order(cls, tn, order):
        tn_copy = tn.copy()
        tree_dic = {}
        edge_list = list(e[1] for e in tn_copy.edges)
        for node in tn_copy.nodes:
            tree_dic[node[1]] = cls(subscripts=''.join([opt_einsum.get_symbol(edge_list.index(e)) for e in tn_copy.network.nodes[node]['edges']]), name=node[1])
        for o in order:
            stn_name, _ = tn_copy.encapsulate(o[0], stn_name=o[1])
            tree_dic[stn_name] = cls(subscripts=''.join([opt_einsum.get_symbol(edge_list.index(e)) for e in tn_copy.network.nodes[(0, stn_name)]['edges']]), name=o[1], children=[tree_dic[i] for i in o[0]])
        return tree_dic['#']

    def __init__(self, subscripts=None, children=None, name=None):
        self.name = name
        self.subs = subscripts
        self.children = children
        if self.children is not None:
            self.all_subscripts = set()
            for c in self.children:
                # c.parent = self
                self.all_subscripts |= set(c.subs)
            self.children = sorted(self.children, key=lambda x: len(x.subs))
        else: self.all_subscripts = self.subs

    @property
    def subscripts(self):
        if self.children is None:
            return []
        res = [[[c.subs for c in self.children], self.subs]]
        return res

    @property
    def full_subscripts(self):
        res = []
        if self.children is None: return res
        for c in self.children[::-1]:
            res += c.full_subscripts
        return res + self.subscripts

    @property
    def order(self):
        if self.children is None: return []
        return [[[c.name for c in self.children], self.name]]

    @property
    def full_order(self):
        res = []
        if self.children is None: return res
        assert len(self.children) <= 2 and self not in self.children
        for c in self.children[::-1]:
            res += c.full_order
        return res + self.order

    @property
    def step_cost(self):
        if self.children is None: return ContractionCost(1, 1)
        else: return ContractionCost(2**len(self.all_subscripts), 2**len(self.subs))

    @property
    def total_cost(self):
        res = self.step_cost
        if self.children is not None:
            for c in self.children:
                res += c.total_cost
        return res

    def flip_branch(self):
        flg = False
        if len(self.subs) >= CONTRACTION_TREE_PARAMS['stem_size'] and\
           self.children is not None and\
           self.children[1].children is not None:
            stem0 = self.children[1].children[1]
            stem1 = self.children[1]
            branch0 = self.children[1].children[0]
            branch1 = self.children[0]
            imag_stem1_remain = ''.join(set(stem0.subs).union(branch1.subs).intersection(set(self.subs).union(branch0.subs)))
            if len(imag_stem1_remain) < len(stem1.subs) or (len(imag_stem1_remain) == len(stem1.subs) and branch0.subs > branch1.subs):
                flg = True
                imag_stem1 = ContractionTree(children=[branch1, stem0], subscripts=imag_stem1_remain, name=stem1.name)
                self.children = [branch0, imag_stem1]
                self.all_subscripts = set(imag_stem1.subs).union(branch0.subs)
        if self.children is not None:
            for c in self.children:
                flg |= c.flip_branch()
        return flg

    def branch_info(self, c=True):
        res = []
        if c is True:
            if self.children is None:
                return res
            if max(len(self.children[0].subs), len(self.children[1].subs)) >= CONTRACTION_TREE_PARAMS['stem_size']:
                res = [[self, self.subs]]
                c = False
                for cc in self.children:
                    res.append([[cc, cc.subs]]+cc.branch_info(c))
            else:
                return self.children[0].branch_info() + self.children[1].branch_info()

        if len(self.subs) >= CONTRACTION_TREE_PARAMS['stem_size']:
            if self.children is not None:
                m = set(self.children[0].subs).difference(self.children[1].subs)
                n = set(self.children[0].subs).intersection(self.children[1].subs)
                res.append([self.children[0],  ''.join(n), ''.join(m), self.children[1]])
            if self.children is not None:
                for cc in self.children:
                    res += cc.branch_info(c)
        return res

    def branches_to_digraph(self, branches):
        k = branches[0]
        if len(branches) <= 3:
            return None
        g = networkx.MultiDiGraph()
        g.add_node(0, branch=k[0])
        edges_dic = {a: 0 for a in k[1]}
        self.inter_names = []
        ll = len(branches) - 1
        for i, ss in enumerate(branches):
            if len(ss) > 2:
                self.inter_names.append(ss[3].name)
                g.add_node(i+1, branch=ss[0])
                for n in ss[2]:
                    g.add_edge(i+1 if i < ll-1 else -1, edges_dic[n], name=n)
                    del edges_dic[n]
                for n in ss[1]:
                    edges_dic[n] = i+1
        g.nodes[-1]['branch'] = ss[3]
        for n in edges_dic:
            g.add_edge(-1, edges_dic[n], name=n)
        return g

    def reconstruct_order(self, g):
        curr_node = g.nodes[-1]['branch']
        for node in [i for i in list(g.nodes) if i != -1][:0:-1]:
            nn = g.nodes[node]['branch']
            subs = set(nn.subs).difference(curr_node.subs) | set(curr_node.subs).difference(nn.subs)
            new_node = ContractionTree(children=[nn, curr_node], subscripts=subs, name=self.inter_names[0])
            del self.inter_names[0]
            curr_node = new_node
        g.nodes[0]['branch'].children = curr_node.children

    def merge_nodes(self, width):
        br = self.branch_info()
        if len(br[1]) >= len(br[2]):
            brb = [br[2], br[1]]
        else:
            brb = [br[1], br[2]]
        m0, m1 = set(brb[0][0][1]), set(brb[1][0][1])
        i = 0
        while (len(m0) + len(m1) - 2*len(m0.intersection(m1)) <= 12) and i < len(brb):
            i += 1
            m0 = m0.union(brb[0][i][1]).difference(brb[0][i][2])
        if i > 1:
            k1 = brb[0][i][0]
            p1 = brb[1][0][0]
            br[0][0].children = [k1, brb[0][0][0]]
            brb[0][i-1][3].children = [brb[0][i][3], p1]
            brb[0][i] = [brb[0][i][3], ''.join(m0)]

            brb[0] = brb[0][i:]
        # assert False
        for brk in brb:
            g = self.branches_to_digraph(brk)
            if g is None:
                continue
            while True:
                a = networkx.Graph()
                g_list = list(g.nodes)
                for i, u in enumerate(g_list):
                    if u == 0 or u == -1 or i >= len(g_list) - 2:
                        continue
                    vs = [list(g_list)[i+1] if list(g_list)[i+1]!= -1 else list(g_list)[-2]]
                    vv = sorted([i for i in list(g.predecessors(u)) if i != -1])
                    if len(vv) > 0 and vv[0] != vs[0]:
                        vs.append(vv[0])
                        # if v == -1: continue
                    for v in vs:
                        if g.in_degree(u) <= width and g.out_degree(v) <= width and g.out_degree(u)<=g.in_degree(u):
                            ustr = g.nodes[u]['branch'].subs
                            vstr = g.nodes[v]['branch'].subs
                            kstr = set(ustr).symmetric_difference(vstr)
                            if 2**len(ustr) + 2**len(vstr) >= 2**(len(kstr)) / CONTRACTION_TREE_PARAMS['k']:
                                a.add_edge(u, v, weight = 2**(-len(kstr)))

                matching = max_weight_matching(a)
                if len(matching) == 0:
                    break
                for u, v in matching:
                    u, v = max(u, v), min(u, v)
                    bu = g.nodes[u]['branch']
                    bv = g.nodes[v]['branch']
                    subs = ''.join(set(bu.subs).difference(bv.subs) | set(bv.subs).difference(bu.subs))
                    for node in g.successors(v):
                        if node != u:
                            for e in g[v][node]:
                                g.add_edge(u, node, name=g[v][node][e]['name'])
                    for node in g.predecessors(v):
                        if node != u:
                            for e in g[node][v]:
                                g.add_edge(node, u, name=g[node][v][e]['name'])
                    name = self.inter_names[0]
                    del self.inter_names[0]
                    g.nodes[u]['branch'] = ContractionTree(children=[bu, bv], name=name, subscripts=subs)
                    g.remove_node(v)

                a = networkx.Graph()
                g_list = list(g.nodes)
                for i, u in enumerate(g_list):
                    if u == 0 or u == -1 or i <= 1: continue
                    vs = [g_list[i-1]]
                    vv = sorted([i for i in list(g.successors(u)) if i != 0])
                    if len(vv) > 0 and vv[-1] != vs[0]:
                        vs.append(vv[-1])
                    for v in vs:
                        # if v == 0: continue
                        if g.out_degree(u) <= width and g.in_degree(v) <= width and g.in_degree(u)<=g.out_degree(u):
                            ustr = g.nodes[u]['branch'].subs
                            vstr = g.nodes[v]['branch'].subs
                            kstr = set(ustr).symmetric_difference(vstr)
                            if 2**len(ustr) + 2**len(vstr) >= 2**(len(kstr)) / CONTRACTION_TREE_PARAMS['k']:
                                a.add_edge(u,v, weight= 2**(-len(kstr)))

                matching = max_weight_matching(a)
                if len(matching) == 0:
                    break
                for u, v in matching:
                    u,v = min(u,v), max(u,v)
                    bu = g.nodes[u]['branch']
                    bv = g.nodes[v]['branch']
                    subs = ''.join(set(bu.subs).difference(bv.subs) | set(bv.subs).difference(bu.subs))
                    for node in g.successors(v):
                        if node != u:
                            for e in g[v][node]:
                                g.add_edge(u, node, name=g[v][node][e]['name'])
                    for node in g.predecessors(v):
                        if node != u:
                            for e in g[node][v]:
                                g.add_edge(node, u, name=g[node][v][e]['name'])
                    name = self.inter_names[0]
                    del self.inter_names[0]
                    g.nodes[u]['branch'] = ContractionTree(children=[bu, bv], name=name, subscripts=subs)
                    g.remove_node(v)
            self.reconstruct_order(g)

    def patch(self, width=5):
        self.merge_nodes(width)
        return self.full_order



from setuptools import setup, find_packages
import os.path

this_dir = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(this_dir, "README.md"), "r") as fh:
    long_description = fh.read()

zip_safe = True
try:
    from Cython.Build import cythonize
    zip_safe = False
except ImportError:
    print('Installing QDP without Cython...')
    def cythonize(*args, **kwargs):
        return []

setup(
    name='acqdp',
    version='0.1.1',
    description='Alibaba Cloud Quantum Development Platform',
    long_description = long_description,
    long_description_content_type='text/markdown',
    author='Alibaba Quantum Lab',
    author_email='aql_software@alibabacloud.com',
    license='MIT',
    url='https://github.com/alibaba/acqdp',
    packages=find_packages(include=['acqdp*']),
    package_data={'acqdp': ['*.ini', '*.json']},
    include_package_data=True,
    ext_modules=cythonize('acqdp/utility/*.pyx'),
    zip_safe=zip_safe,
    install_requires=[
        'numpy',
        'scipy',
        'networkx',
        'numexpr',
        'matplotlib',
        'opt_einsum',
        'kahypar',
        'cma',
        'jax',
        'jaxlib',
        'tqdm'
    ],
    python_requires='>=3.7'
)



import subprocess
import os
import copy
import sys
import string
import random
import networkx as nx
import itertools
import numpy as np
import opt_einsum
import json
from functools import cmp_to_key
from acqdp.tensor_network.contraction import ContractionScheme
from acqdp.tensor_network.local_optimizer import defaultOrderResolver

if sys.version_info < (3, 0):
    sys.stdout.write("Sorry, requires Python 3.x, not Python 2.x\n")
    sys.exit(1)


class OrderFinder:
    """
    :class:`OrderFinder` class is dedicated to finding a contraction scheme corresponding to a given tensor network structure.
    The main method of an :class:`OrderFinder` is `find_order`, which takes a `TensorNetwork` as input, and yields a generator of `ContractionScheme`.
    The base class offers preliminary order finding schemes. For more advanced hypergraph-based approach, use :class:`KHPOrderFinder`. For finding contraction schemes with sliced edges, use :class:`SlicedOrderFinder`.

    :ivar order_method: 'default' order contracts the tensor one by one by order; 'vertical' order contracts the tensor based on the vertical ordering. When the tensor network is from a quantum circuit, the vertical order corresponds to the tensor network contraction where the tensors connected to a qubit is first contracted together and then merged to a main one according to a given order of the qubits.
    """
    def __init__(self,
                 order_method='default',
                 **kwargs):
        self.order_method = order_method

    def find_order(self, tn, **kwargs):
        """
        Find a contraction scheme for the input tensor network subject to the constraints given in the :class:`OrderFinder`.

        :param tn: A tensor network for which the contraction scheme is to be determined.
        :type tn: :class:`TensorNetwork`
        :yields: A :class:`ContractionScheme` containing the pairwise contraction order, a list of edges to be sliced, and optionally the total contraction cost.

        """
        tn = tn._expand_and_delta()
        if self.order_method == 'default':
            a = lambda b: b
        elif self.order_method == 'vertical':
            qubit_order = kwargs.get('qubit_order', sorted(set(b[1] for b in tn.nodes_by_name)))
            a = lambda b: (qubit_order.index(b[1]), b[0], b[2:])
        else:
            raise ValueError("order method not implemented")
        try:
            nodes_list = sorted(tn.nodes_by_name, key=a)
        except TypeError:
            nodes_list = sorted(tn.nodes_by_name, key=lambda x: str(x))
        o = []
        if len(nodes_list) > 1:
            k = nodes_list[0]
            for i in range(len(nodes_list) - 1):
                new_k = ('#', i)
                o.append([[nodes_list[i + 1], k], new_k])
                k = new_k
            if len(o) > 0:
                o[-1][-1] = '#'
        res = defaultOrderResolver.order_to_contraction_scheme(tn, o)
        while True:
            yield res


class SlicedOrderFinder(OrderFinder):
    """
    :class: `SlicedOrderFinder` finds a sliced contraction scheme based on unsliced contraction schemes found by its base order finder.
    :ivar base_order_finder: The base order finder of the sliced order finder, from which the `SlicedOrderFinder` fetches contraction schemes and do slicing on it.
    :ivar slicer: The slicing algorithm acting upon the contraction schemes.
    :ivar num_candidates: Number of unsliced contraction schemes to feed to the slicer at a time. Set to 20 by default.
    """
    def __init__(self,
                 base_order_finder={'order_finder_name': 'khp'},
                 slicer={'slicer_name': 'default'},
                 num_candidates=20,
                 **kwargs):
        self.base_order_finder = base_order_finder
        self.num_candidates = num_candidates
        from acqdp.tensor_network.order_finder import get_order_finder
        from acqdp.tensor_network.slicer import get_slicer
        self.base_order_finder = get_order_finder(**base_order_finder)
        self.slicer = get_slicer(**slicer)

    def find_order(self, tn, **kwargs):
        tn = tn._expand_and_delta()
        order_gen = self.base_order_finder.find_order(graph=tn)
        next(order_gen)
        while True:
            res = self.slicer.slice(tn, order_gen)
            yield res


def get_order_finder(**kwargs):
    from acqdp.tensor_network.kahypar_order_finder import KHPOrderFinder
    order_finders = {
        'khp': KHPOrderFinder,
        'default': OrderFinder,
        'sliced': SlicedOrderFinder
    }
    order_finder_name = kwargs.get('order_finder_name', 'default')
    return (order_finders[order_finder_name])(
        **kwargs.get('order_finder_params', {}))



from .tensor_network import TensorNetwork
from .tensor_sum import TensorSum
from .tensor import Tensor
from .tensor_view import TensorView
from .tensor_valued import TensorValued, normalize, conjugate, transpose
from .order_finder import OrderFinder, SlicedOrderFinder
from .kahypar_order_finder import KHPOrderFinder
from .slicer import Slicer, MPSlicer

from .contraction import (ContractionCost, ContractionScheme,
                                         ContractionTask)

from .local_optimizer import (LocalOptimizer, OrderResolver,
                                             defaultOrderResolver)

from .compiler import Compiler

from .contractor import Contractor

from .order_finder import get_order_finder

from .slicer import get_slicer

__all__ = [
    'TensorNetwork', 'Tensor', 'TensorView', 'TensorSum', 'TensorValued',
    'ContractionCost', 'ContractionScheme', 'ContractionTask', 'LocalOptimizer',
    'OrderResolver', 'defaultOrderResolver', 'Compiler', 'Contractor',
    'get_order_finder', 'get_slicer', 'normalize', 'conjugate', 'transpose',
    'OrderFinder', 'SlicedOrderFinder', 'KHPOrderFinder', 'Slicer', 'MPSlicer'
]



import copy
import numpy
from . import contractor

DTYPE = complex

def conjugate(t) -> 'TensorValued':
    """
    Return the complex conjugation of a `TensorValued` object.

    :returns:  :class:`TensorValued` -- the conjugation of a `TensorValued` object.
    """
    from .tensor import Tensor
    from .tensor_sum import TensorSum
    from .tensor_network import TensorNetwork
    from .tensor_view import TensorView
    if not isinstance(t, TensorValued):
        raise TypeError()
    elif isinstance(t, Tensor):
        if t._data is None:
            return Tensor()
        else:
            return Tensor(numpy.conj(t._data))
    elif isinstance(t, TensorSum):
        ts = t.copy()
        for term in ts.terms_by_name:
            ts.terms_by_name[term] = conjugate(ts.terms_by_name[term])
        return ts
    elif isinstance(t, TensorNetwork):
        tn = t.copy()
        for node in tn.nodes:
            tn.network.nodes[node]['tensor'] = conjugate(tn.network.nodes[node]['tensor'])
        return tn
    elif isinstance(t, TensorView):
        return TensorView(t)

def transpose(t, axes: tuple) -> 'TensorNetwork':
    """
    Return the transposition of a `TensorValued` object.

    :param axes: the transposition on the referred object.
    :type axes: tuple.
    :returns:  :class:`TensorNetwork` -- the transposition of a `TensorValued` object can be readily expressed in terms of a `TensorNetwork` object.
    """
    from .tensor_network import TensorNetwork
    tn = TensorNetwork()
    tn.add_node(tensor=t, is_open=True)
    tn.open_edges = [tn.open_edges[i] for i in axes]
    return tn

def normalize(tv: 'TensorValued') -> 'TensorValued':
    """
    Return a copy with unit Frobenius norm.

    :param tv: The :class:`TensorValued` object to normalize.
    :type tv: :class:`TensorValued`.
    :returns:  :class:`TensorValued` -- the normalized :class:`TensorValued` object.
    """
    return tv * (1/numpy.sqrt(tv.norm_squared))

class TensorValued(object):
    """
    Interface for all :class:`TensorValued` objects, including :class:`Tensor`, :class:`TensorNetwork`, :class:`TensorSum` and :class:`TensorView`.

    :ivar identifier: unique identifier for each :class:`TensorValued` object.
    :ivar dtype: A :class:`TensorValued` object is homogeneous, and contains elements described by a dtype object.
    """
    id_count = -1

    def __new__(cls, *args, **kwargs):
        TensorValued.id_count += 1
        instance = object.__new__(cls)
        return instance

    def __init__(self, dtype: type = DTYPE) -> None:
        self.identifier = TensorValued.id_count
        self.dtype = dtype

    @property
    def shape(self):
        """
        The common property of all :class:`TensorValued` classes, indicating whether the bond dimensions for a tensor valued object. A tensor valued object is semantically a multi-dimensionally array, and its dimensions can be expressed as a tuple of integers. The tuple is called the shape of the tensor, whereas the length of the tuple is the rank of the tensor.

        In ACQDP, undetermined tensors and undetermined dimensions are allowed. In the former case, the shape of the tensor will return `None`; in the latter case, some of the bond_dimensions appearing in the tuple could be `None`.

        :returns: :class:`tuple` or None
        :raises: NotImplementedError, ValueError
        """
        raise NotImplementedError()

    @property
    def is_valid(self) -> bool:
        """
        The common property of all :class:`TensorValued` classes, indicating whether the :class:`TensorValued` object is valid or not.
        In every step of a program, all existing :class:`TensorValued` object must be valid, otherwise an exception should be thrown out.

        :returns: :class:`bool`
        :raises: NotImplementedError
        """
        raise NotImplementedError()

    @property
    def is_ready(self) -> bool:
        """
        The common property of all :class:`TensorValued` classes, indicating whether the current :class:`TensorValued` object is ready for
        contraction, i.e. whether it semantically represents a tensor with a definite value.
        A `TensorValued` object needs to be ready upon contraction, but needs not to be ready throught the construction.

        :returns: :class:`bool`
        :raises: NotImplementedError()
        """
        raise NotImplementedError()

    def __str__(self) -> str:
        return "Type: " + str(type(self))\
               + ", Shape: " + str(self.shape)

    def __repr__(self) -> str:
        return "Id: " + str(self.identifier) + self.__str__()

    def __eq__(self, other) -> bool:
        if self is other:
            return True
        elif isinstance(other, TensorValued):
            if self.shape != other.shape:
                return False
            return numpy.allclose(self.contract(), other.contract())
        else:
            return False

    def __mul__(self, other) -> 'TensorNetwork':
        """
        Tensor product of two tensors. For tensor multiplications, use
        `TensorNetwork` classes to specify how indices are to be contracted.
        By default, addition creates a `TensorNetwork` object with two operands as two components.

        :returns: :class:`TensorNetwork`
        """
        from .tensor_network import TensorNetwork
        tn = TensorNetwork()
        tn.add_node(tensor=self, is_open=True)
        tn.add_node(tensor=other, is_open=True)
        return tn

    __rmul__ = __mul__

    def __add__(self, other: 'TensorValued') -> 'TensorValued':
        """
        Addition of two `TensorValued` objects.
        By default, addition creates a `TensorSum` object with two operands as two components. The two operands of the addition must have compatible shapes.
        """
        from .tensor_sum import TensorSum
        tl = TensorSum()
        tl.add_term(tensor=self)
        tl.add_term(tensor=other)
        return tl

    def __neg__(self) -> 'TensorValued':
        """Negation of a `TensorValued` object."""
        return -1 * self

    def __sub__(self, other: 'TensorValued') -> 'TensorValued':
        """Subtraction of `TensorValued` objects."""
        return self + (-other)

    def __invert__(self) -> 'TensorValued':
        """Syntactic sugar for conjugation."""
        return conjugate(self)

    def __mod__(self, axes: tuple):
        """Syntactic sugar for transposition."""
        return transpose(self, axes)

    @property
    def norm_squared(self):
        """
        Square of Frobenius norm of the :class:`TensorValued` object.
        """
        from .tensor_network import TensorNetwork
        tn = TensorNetwork()
        tn.add_node(tensor=self)
        tn.add_node(tensor=~self)
        return tn.contract().real

    def fix_index(self, index, fix_to=0) -> 'TensorValued':
        """
        Fix the given index to the given value. The result :class:`TensorValued` object would have the same type as the original one, with rank 1 smaller
        than the original.

        :param index: The index to fix.
        :type index: :class:`int`
        :param fix_to: the value to assign to the given index.
        :type fix_to: :class:`bool`
        :returns:  :class:`TensorValued` -- The :class:`TensorValued` object after fixing the given index.
        :raises: NotImplementedError
        """
        raise NotImplementedError()

    def expand(self):
        """
        Expand nested tensor network structures in the :class:`TensorValued` object if there is any.
        :returns: :class:`TensorValued`.
        """
        return self

    def contract(self, **kwargs) -> numpy.ndarray:
        """
        Evaluate the :class:`TensorValued` object to a :class:`numpy.ndarray`.

        :returns: :class:`numpy.ndarray`
        :raises: NotImplementedError
        """
        raise NotImplementedError()

    def cast(self, dtype):
        """
        Cast the tensor valued object to a new underlying dtype.
        """
        self.dtype = dtype

    def copy(self) -> 'TensorValued':
        """Make a copy of the current object. Data is duplicated only when
        necessary."""
        raise NotImplementedError()

    def __deepcopy__(self, memo) -> 'TensorValued':
        """Make a deepcopy of the current object with all data duplicated."""
        raise NotImplementedError()



import json
import numpy as np
import copy
import opt_einsum
from acqdp.tensor_network.contractor import defaultContractor


class ContractionCost:
    """
    :class:`ContractionCost` is a wrapper for cost quantities related to a tensor network contraction.

    :ivar s: Total number of floating point operations, as a proxy for the time complexity.
    :ivar t: Size of largest intermediate tensor, as a proxy for the space complexity.
    :ivar k: Number of indices sliced, as a proxy for the parallelism
    """
    def __init__(self, s=0, t=0, k=0):
        self.s = s
        self.t = t
        self.k = k

    def __add__(self, other):
        return ContractionCost(self.s + other.s, max(self.t, other.t))

    def __iadd__(self, other):
        self.s = self.s + other.s
        self.t = max(self.t, other.t)
        return self

    def __eq__(self, other):
        return self.s * 2**self.k == other.s * 2**other.k and self.t == other.t

    def __lt__(self, other):
        return self.s * 2**self.k < other.s * 2**other.k

    def __str__(self):
        return "cost = {}, cw = {}, num_slice = {}".format(
            np.log10(float(self.s * 2**self.k)), np.log2(float(self.t)), self.k)


class ContractionScheme:
    """
    :class:`ContractionScheme` is an intermediate representation of a tensor network contraction, containing the pairwise contraction orders, the hyperedges to be sliced, and the cost of the contraction. A :class:`ContractionScheme` does not contain tensor data.

    :ivar order: The pairwise sequential contraction order.
    :ivar slices: The hyperedges to be sliced for parallelism.
    :ivar cost: :class:`ContractionCost` -- The time / space cost and number of sub-processes of the tensor network contraction.
    """

    def __init__(self, order, slices=None, cost=None):
        self.order = order
        if slices is None:
            slices = []
        self.slices = slices
        if cost is None:
            cost = ContractionCost()
        self.cost = cost

    def dump(self, f):
        """
        Dump the :class:`ContractionScheme` into a json file.

        :param f: The file to dump the :class:`ContractionScheme` to.
        """
        json.dump(
            {
                'contractions': [l for l in self.order],
                'slices': [l for l in self.slices],
                'cost': [self.cost.s, self.cost.t, self.cost.k]
            },
            f,
            indent=4)

    @classmethod
    def load(cls, f):
        """
        Load a :class:`ContractionScheme` from an existing file.

        :param f: The file to load the :class:`ContractionScheme` from.

        :returns: :class:`ContractionScheme`
        """
        def tuplize(l):
            if isinstance(l, list):
                return tuple(tuplize(x) for x in l)
            return l
        res = json.load(f)
        return cls([[[tuplize(l0), tuplize(l1)],
                     tuplize(l)] for (l0, l1), l in res['contractions']],
                   [tuplize(l) for l in res['slices']],
                   ContractionCost(*res['cost']))


class OrderCounter:
    @property
    def cnt(self):
        self.k += 1
        return ('#', self.k)

    def __init__(self, order=None):
        self.k = 0
        if order is not None:
            for o in order:
                for k in o[0]:
                    if isinstance(k, tuple) and len(k) == 2 and k[0] == '#':
                        self.k = max(self.k, k[1])

    def __copy__(self):
        oc = OrderCounter([[[('#', self.k)], '#']])
        return oc


class ContractionTask:
    """
    :class:`ContractionTask` is post-compilation intermediate representation for tensor network contraction, with all the information necessary to carry out a tensor network contraction, including initial tensor data and stepwise instruction of the contraction.
    """
    def __init__(self,
                 output,
                 inputs=None,
                 commands=None,
                 data=None,
                 length=1,
                 shape=None,
                 fix_dict=None,
                 open_edges=None,
                 sub_outputs=None,
                 cnt=0,
                 dtype=complex):
        if commands is None:
            self.commands = []
        else:
            self.commands = commands
        self.output = output
        if inputs is None:
            self.inputs = []
        else:
            self.inputs = inputs
        if data is None:
            self.data = {}
        else:
            self.data = data
        if shape is None:
            self.shape = ()
        else:
            self.shape = shape
        if fix_dict is None:
            fix_dict = {}
        self.fix_dict = fix_dict
        if sub_outputs is None:
            sub_outputs = []
        self.sub_outputs = sub_outputs
        if open_edges is None:
            open_edges = []
        self.open_edges = open_edges
        self.cnt = cnt
        if cnt > 0:
            self.preprocess = False
        length = 1
        for k in self.fix_dict:
            length *= len(k[1])
        self.length = length
        self.dtype = np.dtype(dtype)


    def _merge(self, res_dic):
        if self.length == 1 or len(self.shape) == 0 or self.open_edges is None:
            return sum([res_dic[k] for k in res_dic])
        for i in res_dic:
            res_type = np.array(res_dic[i]).dtype
        lst_id = [i for i in self.open_edges if i not in self.sub_outputs]
        lst = [
            i for i in range(len(self.shape))
            if self.open_edges[i] not in self.sub_outputs
        ] + [self.open_edges.index(i) for i in self.sub_outputs]

        res = np.transpose(
            np.zeros(self.shape, dtype=res_type),
            lst)
        for k in res_dic:
            a = k
            ax = {}
            for l in self.fix_dict:
                if l[0][1] not in self.sub_outputs:
                    ax[l[0][1]] = a % len(l[1])
                    a //= len(l[1])
            ls = []
            for l in lst_id:
                ls.append(ax[l])

            res[tuple(ls)] += res_dic[k]
        res = np.transpose(
            res, [lst.index(i) for i in range(len(self.shape))])  # lst)
        return res

    def update_fix(self, tn):
        """
        Update the tensor index fixes according to the input tensor network. Use when the contraction task corresponds to the same tensor network with different fix configurations. One example of the use case is when the contraction task is for an entry of the amplitude of a quantum circuit, whereas the tensor network represents the same quantum circuit with a different amplitude entry.

        :param tn: The tensor network to update the fixes to.
        :type tn: :class:`TensorNetwork`.
        """
        new_dict = {}
        for k in self.fix_dict:
            if 'fix_to' in tn.network.nodes[k[0]]:
                c = tn.network.nodes[k[0]]['fix_to']
                new_k = (k[0], tuple(c))
                new_dict[new_k] = self.fix_dict[k]
                new_dict[new_k][0] = list(c)[0]
            else:
                new_dict[k] = self.fix_dict[k]
        self.fix_dict = new_dict
        self.preprocess = False

    def _load_data(self):
        try:
            dilist = sorted(list(self.inputs))
        except TypeError:
            dilist = sorted(list(self.inputs), key=str)
        try:
            ddlist = sorted(list(self.data))
        except TypeError:
            ddlist = sorted(list(self.data), key=str)

        for d in range(len(self.inputs)):
            di = dilist[d]
            dd = ddlist[d]
            if isinstance(self.data[dd], ContractionTask):
                self.inputs[di][0] = (0, np.array(self.data[dd].execute()))
            else:
                self.inputs[di][0] = (0, np.array(self.data[dd]))
        self.preprocess = False
        return self

    def set_data(self, data):
        """
        Update the data in the contractions. Use when a connectivity of a tensor network remains unaltered but the tensor data is changed.

        :param data: a dictionary of tensor data, with keys being their names in the tensor network, and values the corresponding tensor data.
        :type data: dict
        """
        self.data = {d: np.array(data[d], dtype=self.dtype) for d in data}

    def cast(self, dtype):
        """
        Cast the contraction to a different type. All input tensors will be cast to the type, and therefore all steps of contraction will be of the type, together with the final result.

        :param dtype: The dtype to cast the computations to.
        :type dtype: `str` or `type`
        """
        self.dtype = np.dtype(dtype)
        self.preprocess = False
        self.set_data(self.data)

    def __getitem__(self, idx):
        """
        Get a subtask induced by slicing as another :class:`ContractionTask`.

        :param idx: Index of the subtask.
        :type idx: int

        :returns: :class:`ContractionTask` -- The contraction task corresponding to the subtask specified by the index.
        """
        if not self.preprocess:
            self._load_data()
            self.preprocess = True
            defaultContractor._execute(self, cnt=self.cnt)
        self_copy = copy.copy(self)
        for k in self_copy.fix_dict:
            a = idx % len(k[1])
            idx = idx // len(k[1])
            self_copy.fix_dict[k][0] = k[1][a]
        self_copy.commands = self.commands[self.cnt:]
        self_copy.fix_dict = {}
        self_copy.length = 1
        self_copy.cnt = 0
        self_copy.open_edges = None
        self_copy.data = {}
        self_copy.inputs = {}
        return self_copy

    def execute(self, **kwargs):
        """
        Execute a :class:`ContractionTask`. Equivalent to :meth:`Contractor(**kwargs).execute(self)`.

        """
        from acqdp.tensor_network.contractor import Contractor
        return Contractor(**kwargs.get('contractor_params', {})).execute(self)
